use std::collections::BTreeMap;

use anyhow::{bail, Error};
use cozo::{DataValue, DbInstance, Num, ScriptMutability};
use serde_with::chrono::{DateTime, Utc};
use ubisync_lib::types::PotId;

use crate::{api::v0::app::{AppId, App}, run_query};

pub fn add(
    db: &DbInstance,
    id: &AppId,
    last_access: &DateTime<Utc>,
    name: &str,
    description: &str,
    app_type: &str,
) -> anyhow::Result<()> {
    let params = vec![
        (
            "id",
            DataValue::Str(serde_json::to_string(&id).unwrap().into()),
        ),
        (
            "last_access",
            DataValue::Num(Num::Int(last_access.timestamp())),
        ),
        ("name", DataValue::Str(name.into())),
        ("description", DataValue::Str(description.into())),
        ("app_type", DataValue::Str(app_type.into())),
        ("default_pot", DataValue::Null),
    ];

    match run_query!(
        &db,
        ":insert apps {id => last_access, app_type, name, description, default_pot}",
        params,
        cozo::ScriptMutability::Mutable
    ) {
        Ok(_) => Ok(()),
        Err(report) => bail!(report),
    }
}

pub fn exists(db: &DbInstance, id: &AppId) -> anyhow::Result<bool> {
    let mut params = BTreeMap::new();
    params.insert(
        "id".to_string(),
        DataValue::Str(serde_json::to_string(&id)?.into()),
    );

    let result = db.run_script(
        "?[name] := *apps[$id, last_access, app_type, name, description, default_pot]",
        params,
        cozo::ScriptMutability::Immutable,
    );

    if let Ok(rows) = result {
        return Ok(rows.rows.len() == 1);
    }

    Err(Error::msg("Could not check whether app is registered"))
}

pub fn get(db: &DbInstance, id: &AppId) -> anyhow::Result<App> {
    let mut params = BTreeMap::new();
    params.insert(
        "id".to_string(),
        DataValue::Str(serde_json::to_string(&id)?.into()),
    );

    let result = db.run_script(
        "?[last_access, app_type, name, description, default_pot] := *apps[$id, last_access, app_type, name, description, default_pot]",
        params,
        cozo::ScriptMutability::Immutable,
    );

    if let Ok(rows) = result {
        if let Some(firstrow) = rows.rows.first() {
            if let [DataValue::Num(Num::Int(ts)), DataValue::Str(app_type), DataValue::Str(name), DataValue::Str(desc), default_pot] = firstrow.as_slice() {
                let last_access = DateTime::from_timestamp(ts.to_owned(), 0).ok_or(Error::msg("Failed to deserialize timestamp"))?;
                let pot_id = match default_pot {
                    DataValue::Str(dpid) => Some(serde_json::from_str(dpid)?),
                    _ => None,
                };
                return Ok(App {id: id.clone(), app_type: app_type.to_string(), last_access, name: name.to_string(), description: desc.to_string(), default_pot: pot_id});
            }
        }
    }

    Err(Error::msg("Could not find app"))
}

pub fn set_default_pot(db: &DbInstance, id: &AppId, pot: &PotId) -> anyhow::Result<()> {
    let params = vec![
        ("id", DataValue::Str(serde_json::to_string(id)?.into())),
        ("default_pot", DataValue::Str(serde_json::to_string(pot)?.into()))
    ];

    match run_query!(
        &db,
        ":update apps {id => default_pot}",
        params,
        ScriptMutability::Mutable
    ) {
        Ok(_) => Ok(()),
        Err(report) => bail!(report),
    }
}

pub fn create_pot(db: &DbInstance, pot_id: &PotId, app_id: &AppId, app_type: &str) -> anyhow::Result<()> {
    let params = vec![
        ("pot_id", DataValue::Str(serde_json::to_string(pot_id)?.into())),
        ("app_id", DataValue::Str(serde_json::to_string(app_id)?.into())),
        ("app_type", DataValue::Str(app_type.into())),
    ];

    match run_query!(
        &db,
        ":insert pots {pot_id => app_type}
         :insert pot_memberships {pot_id => app_id}",
        params,
        ScriptMutability::Mutable
    ) {
        Ok(_) => Ok(()),
        Err(e) => bail!(e),
    }
}